import os
from re import sub as re_sub


def clean_player_id(id:str) -> str:
    """
    cleans player id and returns handle
    """
    return id[id.find(' ')+1:-1]

def clean_entity_id(id:str) -> str:
    """
    cleans entity id and returns it
    """
    return re_sub(r'C\[([0-9]+) +?([a-zA-Z_0-9]+)\]', r'\2 \1', id).replace('_', ' ')

def get_entity_num(id:str) -> int:
    """
    gets entity number from entity id
    """
    if not id.startswith('C['):
        return -1
    try:
        return int(re_sub(r'C\[([0-9]+) +?([a-zA-Z_0-9]+)\]', r'\1', id))
    except ValueError:
        return int(re_sub(r'C\[([0-9]+) +?([a-zA-Z_0-9]+)\]_WCB', r'\1', id))
    except TypeError:
        return -1
    
def compensate_text(text:str) -> str:
    """
    Unescapes various characters not correctly represented in combatlog files

    Parameters:
    - :param text: str -> text to be cleaned

    :return: cleaned text
    """
    text = text.replace('â€“', '–')
    text = text.replace('Ãœ', 'Ü')
    text = text.replace('Ã¼', 'ü')
    text = text.replace('ÃŸ', 'ß')
    text = text.replace('Ã¶', 'ö')
    text = text.replace('Ã¤', 'ä')
    text = text.replace('â€˜', "'")
    return text

def format_path(path: str):
    path = path.replace(chr(92), '/')
    if path[1] == ':' and path[0] >= 'a' and path[0] <= 'z':
        path = path[0].capitalize() + path[1:]
    if path[-1] != '/':
        if os.path.isdir(path):
            path += '/'
    return path

def format_data(el, integer=False) -> str:
    """
    rounds floats and ints to 2 decimals and sets 1000s seperators, ignores string values
    
    Parameters:
    - :param el: value to be formatted
    - :param integer: rounds numbers to zero decimal places when True (optional)
    """
    if isinstance(el, (int, float)):
        if not integer: return f'{el:,.2f}'
        else: return f'{el:,.0f}'
    elif isinstance(el, str):
        el = el.replace('â€“', '–')
        el = el.replace('Ãœ', 'Ü')
        el = el.replace('Ã¼', 'ü')
        el = el.replace('ÃŸ', 'ß')
        el = el.replace('Ã¶', 'ö')
        el = el.replace('Ã¤', 'ä')
        el = el.replace('â€˜', "'")
        return el
    else:
        return str(el)
    
def format_datetime_str(datetime: str) -> str:
    """
    Formats datetime string into datetime to be displayed.

    Parameters:
    - :param datetime: datetime string; for example "2022-02-23T14:00:27.100000Z"

    :return: formated datetime -> "2022-02-23 14:00:27"
    """
    parts = datetime[:-1].split(':')
    seconds = int(float(parts[-1]))
    return f'{parts[0].replace("T", " ")}:{parts[1]}:{seconds:02d}'



from types import FunctionType, BuiltinFunctionType, MethodType

from PySide6.QtWidgets import QPushButton, QFrame, QLabel, QTreeView, QHeaderView, QTableView
from PySide6.QtWidgets import QSizePolicy, QAbstractItemView, QMessageBox, QComboBox
from PySide6.QtWidgets import QVBoxLayout, QHBoxLayout
from PySide6.QtCore import Qt, QSize

from .style import get_style_class, get_style, merge_style, theme_font

CALLABLE = (FunctionType, BuiltinFunctionType, MethodType)

SMINMIN = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
SMAXMAX = QSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum)
SMAXMIN = QSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Minimum)
SMINMAX = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Maximum)

ATOP = Qt.AlignmentFlag.AlignTop
ARIGHT = Qt.AlignmentFlag.AlignRight
ALEFT = Qt.AlignmentFlag.AlignLeft
ACENTER = Qt.AlignmentFlag.AlignCenter
AVCENTER = Qt.AlignmentFlag.AlignVCenter
AHCENTER = Qt.AlignmentFlag.AlignHCenter

RFIXED = QHeaderView.ResizeMode.Fixed

SMPIXEL = QAbstractItemView.ScrollMode.ScrollPerPixel

def create_button(self, text, style: str = 'button', parent=None, style_override={}, toggle=None):
    """
    Creates a button according to style with parent.

    Parameters:
    - :param text: text to be shown on the button
    - :param style: name of the style as in self.theme or style dict
    - :param parent: parent of the button (optional)
    - :param style_override: style dict to override default style (optional)
    - :param toggle: True or False when button should be a toggle button, None when it should be a normal 
    button; the bool value indicates the default state of the button

    :return: configured QPushButton
    """
    button = QPushButton(text, parent)
    button.setStyleSheet(get_style_class(self, 'QPushButton', style, style_override))
    if 'font' in style_override:
        button.setFont(theme_font(self, style, style_override['font']))
    else:
        button.setFont(theme_font(self, style))
    button.setSizePolicy(SMAXMAX)
    if isinstance(toggle, bool):
        button.setCheckable(True)
        button.setChecked(toggle)
    return button

def create_icon_button(self, icon, style:str='', parent=None, style_override={}):
    """
    Creates a button showing an icon according to style with parent.

    Parameters:
    - :param icon: icon to be shown on the button
    - :param style: name of the style as in self.theme or style dict
    - :param parent: parent of the button (optional)
    - :param style_override: style dict to override default style (optional)

    :return: configured QPushButton
    """
    button = QPushButton('', parent)
    button.setIcon(icon)
    button.setStyleSheet(get_style_class(self, 'QPushButton', style, style_override))
    icon_size = self.theme['s.c']['button_icon_size']
    button.setIconSize(QSize(icon_size, icon_size))
    button.setSizePolicy(SMAXMAX)
    return button

def create_frame(self, parent=None, style='frame', style_override={}, size_policy=None) -> QFrame:
    """
    Creates a frame with default styling and parent

    Parameters:
    - :param parent: parent of the frame (optional)
    - :param style: style dict to override default style (optional)
    - :param size_policy: size policy of the frame (optional)

    :return: configured QFrame
    """
    frame = QFrame(parent)
    frame.setStyleSheet(get_style(self, style, style_override))
    frame.setSizePolicy(size_policy if isinstance(size_policy, QSizePolicy) else SMAXMAX)
    return frame

def create_label(self, text, style:str='', parent=None, style_override={}):
    """
    Creates a label according to style with parent.

    Parameters:
    - :param text: text to be shown on the label
    - :param style: name of the style as in self.theme
    - :param parent: parent of the label (optional)
    - :param style_override: style dict to override default style (optional)

    :return: configured QLabel
    """
    label = QLabel(parent)
    label.setText(text)
    label.setStyleSheet(get_style(self, style, style_override))
    label.setSizePolicy(SMAXMAX)
    if 'font' in style_override:
        label.setFont(theme_font(self, style, style_override['font']))
    else:
        label.setFont(theme_font(self, style))
    return label
    
def create_button_series(self, parent, buttons:dict, style, shape:str='row', seperator:str='', ret=False):
    """
    Creates a row / column of buttons.

    Parameters:
    - :param parent: widget that will contain the buttons
    - :param buttons: dictionary containing button details
        - key "default" contains style override for all buttons (optional)
        - all other keys represent one button, key will be the text on the button; value for the key contains
        dict with details for the specific button (all optional)
            - "callback": callable that will be called on button click
            - "style": individual style override dict
            - "toggle": True or False when button should be a toggle button, None when it should be a normal 
            button; the bool value indicates the default state of the button
            - "stretch": stretch value for the button
            - "align": alignment flag for button
    - :param style: key for self.theme -> default style
    - :param shape: row / column
    - :param seperator: string seperator displayed between buttons (optional)

    :return: populated QVBoxlayout / QHBoxlayout
    """
    if 'default' in buttons:
        defaults = merge_style(self, self.theme[style], buttons.pop('default'))
    else:
        defaults = self.theme[style]

    if shape == 'column':
        layout = QVBoxLayout()
    else:
        shape = 'row'
        layout = QHBoxLayout()
    
    layout.setContentsMargins(0, 0, 0, 0)
    layout.setSpacing(0)

    button_list = []
    
    if seperator != '':
        sep_style = {'color':defaults['color'], 'margin':0, 'padding':0, 'background':'rgba(0,0,0,0)'}
    
    for i, (name, detail) in enumerate(buttons.items()):
        if 'style' in detail:
            button_style = merge_style(self, defaults, detail['style'])
        else:
            button_style = defaults
        toggle_button = detail['toggle'] if 'toggle' in detail else None
        bt = self.create_button(name, style, parent, button_style, toggle_button)
        if 'callback' in detail and isinstance(detail['callback'], CALLABLE):
            bt.clicked.connect(detail['callback'])
        stretch = detail['stretch'] if 'stretch' in detail else 0
        if 'align' in detail:
            layout.addWidget(bt, stretch, detail['align'])
        else:
            layout.addWidget(bt, stretch)
        button_list.append(bt)
        if seperator != '' and i < (len(buttons) - 1):
            sep_label = self.create_label(seperator, 'label', parent, sep_style)
            sep_label.setSizePolicy(SMAXMIN)
            layout.addWidget(sep_label)
    
    if ret: return layout, button_list
    else: return layout

def create_combo_box(self, parent, style:str = 'combobox', style_override: dict = {}) -> QComboBox:
    """
    Creates a combobox with given style and returns it.

    Parameters:
    - :param parent: parent of the combo box
    - :param style: key for self.theme -> default style
    - :param style_override: style dict to override default style

    :return: styled QCombobox
    """
    combo_box = QComboBox(parent)
    combo_box.setStyleSheet(get_style_class(self, 'QComboBox', style, style_override))
    if 'font' in style_override:
        combo_box.setFont(theme_font(self, style, style_override['font']))
    else:
        combo_box.setFont(theme_font(self, style))
    combo_box.setSizePolicy(SMINMAX)
    return combo_box

def resize_tree_table(tree: QTreeView):
    """
    Resizes the columns of the given tree table to fit its contents

    Parameters:
    - :param tree: QTreeView -> tree to be resized
    """
    for col in range(tree.header().count()):
        width = max(tree.sizeHintForColumn(col), tree.header().sectionSizeHint(col)) + 5
        tree.header().resizeSection(col, width)
        
def create_analysis_table(self, parent, widget) -> QTreeView:
    """
    Creates and returns a QTreeView with parent, styled according to widget.

    Parameters:
    - :param parent: parent of the table
    - :param widget: style key for the table

    :return: configured QTreeView
    """
    table = QTreeView(parent)
    table.setStyleSheet(get_style_class(self, 'QTreeView', widget))
    table.setSizePolicy(SMINMIN)
    table.setAlternatingRowColors(True)
    table.setHorizontalScrollMode(SMPIXEL)
    table.setVerticalScrollMode(SMPIXEL)
    table.setSortingEnabled(True)
    table.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
    table.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
    table.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
    table.header().setStyleSheet(get_style_class(self, 'QHeaderView', 'tree_table_header'))
    table.header().setSectionResizeMode(RFIXED)
    #table.header().setSectionsMovable(False)
    table.header().setMinimumSectionSize(1)
    table.header().setSectionsClickable(True)
    table.header().setStretchLastSection(False)
    table.header().setSortIndicatorShown(False)
    table.expanded.connect(lambda: resize_tree_table(table))
    table.collapsed.connect(lambda: resize_tree_table(table))
    return table

def style_table(self, table: QTableView):
    """
    Styles the given table.

    Parameters:
    - :param table: table to be styled
    """
    table.setAlternatingRowColors(self.theme['s.c']['table_alternate'])
    table.setShowGrid(self.theme['s.c']['table_gridline'])
    table.setSortingEnabled(True)
    table.setStyleSheet(get_style_class(self, 'QTableView', 'table'))
    table.setHorizontalScrollMode(SMPIXEL)
    table.setVerticalScrollMode(SMPIXEL)
    table.horizontalHeader().setStyleSheet(get_style_class(self, 'QHeaderView', 'table_header'))
    table.verticalHeader().setStyleSheet(get_style_class(self, 'QHeaderView', 'table_index'))
    table.resizeColumnsToContents()
    table.resizeRowsToContents()
    table.horizontalHeader().setSortIndicatorShown(False)
    table.horizontalHeader().setSectionResizeMode(RFIXED)
    table.verticalHeader().setSectionResizeMode(RFIXED)
    table.setSizePolicy(SMINMIN)

def show_warning(self, title: str, message: str):
    """
    Displays a warning in form of a message box

    Parameters:
    - :param title: title of the warning
    - :param message: message to be displayed
    """
    error = QMessageBox()
    error.setIcon(QMessageBox.Icon.Warning), 
    error.setText(message)
    error.setWindowTitle(title)
    error.setStandardButtons(QMessageBox.StandardButton.Ok)
    error.setWindowIcon(self.icons['oscr'])
    error.exec()



from typing import Callable, Iterable

from PySide6.QtWidgets import QVBoxLayout, QHBoxLayout, QWidget, QTableView, QFrame, QLabel
from pyqtgraph import PlotWidget, BarGraphItem, setConfigOptions, mkPen
import numpy as np

from OSCR import TABLE_HEADER

from .datamodels import OverviewTableModel, SortingProxy
from .widgetbuilder import SMINMIN, AVCENTER, ACENTER, create_frame, create_label, style_table
from .widgets import CustomPlotAxis
from .style import get_style, theme_font

setConfigOptions(antialias=True)

def setup_plot(plot_function: Callable) -> Callable:
    '''
    sets up Plot item and puts it into layout
    '''
    def plot_wrapper(self, data, time_reference=None):
        plot_widget = PlotWidget()
        plot_widget.setAxisItems({'left': CustomPlotAxis('left')})
        plot_widget.setAxisItems({'bottom': CustomPlotAxis('bottom')})
        plot_widget.setStyleSheet(get_style(self, 'plot_widget_nullifier'))
        plot_widget.setBackground(None)
        plot_widget.setMouseEnabled(False, False)
        plot_widget.setMenuEnabled(False)
        plot_widget.hideButtons()
        plot_widget.setDefaultPadding(padding=0)
        left_axis = plot_widget.getAxis('left')
        left_axis.setTickFont(theme_font(self, 'plot_widget'))
        left_axis.setTextPen(color=self.theme['defaults']['fg'])
        bottom_axis = plot_widget.getAxis('bottom')
        bottom_axis.setTickFont(theme_font(self, 'plot_widget'))
        bottom_axis.setTextPen(color=self.theme['defaults']['fg'])

        if time_reference is None:
            legend_data = plot_function(self, data, plot_widget)
        else:
            legend_data = plot_function(self, data, time_reference, plot_widget)

        inner_layout = QVBoxLayout()
        inner_layout.setContentsMargins(0, 0, 0, 0)
        inner_layout.setSpacing(self.theme['defaults']['isp'])
        inner_layout.addWidget(plot_widget)
        if legend_data is not None:
            legend_frame = create_legend(self, legend_data)
            inner_layout.addWidget(legend_frame, alignment=ACENTER)
        frame = create_frame(self, None, 'plot_widget')
        frame.setSizePolicy(SMINMIN)
        frame.setLayout(inner_layout)
        outer_layout = QVBoxLayout()
        outer_layout.setContentsMargins(0, 0, 0, 0)
        outer_layout.addWidget(frame, stretch=11) # if stretch ever needs to be variable, create argument for decorator
        return outer_layout
    return plot_wrapper

def create_overview(self):
    """
    creates the main Parse Overview including graphs and table
    """
    # clear graph frames
    for frame in self.widgets.overview_tab_frames:
        if frame.layout():
            QWidget().setLayout(frame.layout())

    time_data, DPS_graph_data, DMG_graph_data = self.parser1.active_combat.graph_data
    current_table = self.parser1.active_combat.table

    line_layout = create_line_graph(self, DPS_graph_data, time_data)
    self.widgets.overview_tab_frames[1].setLayout(line_layout)

    group_bar_layout = create_grouped_bar_plot(self, DMG_graph_data, time_data)
    self.widgets.overview_tab_frames[2].setLayout(group_bar_layout)

    bar_layout = create_horizontal_bar_graph(self, current_table)
    self.widgets.overview_tab_frames[0].setLayout(bar_layout)

    tbl = create_overview_table(self)
    bar_layout.addWidget(tbl, stretch=4)

@setup_plot
def create_grouped_bar_plot(self, data: dict[str, tuple], time_reference: dict[str, tuple], 
            bar_widget: PlotWidget) -> QVBoxLayout:
    """
    Creates a bar plot with grouped bars.

    Parameters:
    - :param data: dictionary containing the data to be plotted
    - :param time_reference: contains the time values for the data points
    - :param bar_widget: bar widget that will be plotted to (supplied by decorator)

    :return: layout containing the graph
    """
    bottom_axis = bar_widget.getAxis('bottom')
    bottom_axis.unit = 's'
    legend_data = list()

    group_width = 0.18
    player_num = len(data)
    bar_width = group_width / player_num
    relative_bar_positions = np.linspace(0+bar_width/2, group_width-bar_width/2, player_num)
    bar_position_offsets = relative_bar_positions - np.median(relative_bar_positions)

    zipper = zip(data.items(), self.theme['plot']['color_cycler'], bar_position_offsets)
    for (player, graph_data), color, offset in zipper:
        if player in time_reference:
            time_data = np.subtract(time_reference[player], offset)
            bars = BarGraphItem(x=time_data, width=bar_width, height=graph_data, brush=color, pen=None, 
                    name=player)
            bar_widget.addItem(bars)
            legend_data.append((color, player))
    return legend_data

@setup_plot
def create_horizontal_bar_graph(self, table: list[list], bar_widget: PlotWidget) -> QVBoxLayout:
    """
    Creates bar plot from table and returns layout in which the graph was inserted.

    Parameters:
    - :param table: overview table as generated by the parser
    - :param bar_widget: bar widget that will be plotted to (supplied by decorator)

    :return: layout containing the graph (returned by decorator)
    """
    left_axis = bar_widget.getAxis('left')
    left_axis.setTickFont(theme_font(self, 'app'))
    bar_widget.setDefaultPadding(padding=0.01)

    y_annotations = (tuple((index+1, line[0]+line[1]) for index, line in enumerate(table)),)
    bar_widget.getAxis('left').setTicks(y_annotations)
    x = tuple(line[3] for line in table)
    y = tuple(range(1, len(x) + 1))
    bar_widget.setXRange(0, max(x) * 1.05, padding=0)
    bars = BarGraphItem(x0=0, y=y, height=0.75, width=x, brush=self.theme['defaults']['mfg'], pen=None)
    bar_widget.addItem(bars)

@setup_plot
def create_line_graph(self, data: dict[str, tuple], time_reference: dict[str, tuple], 
            graph_widget: PlotWidget) -> QVBoxLayout:
    """
    Creates line plot from data and returns layout that countins the plot. 

    Parameters:
    - :param data: dictionary containing the data to be plotted
    - :param time_reference: contains the time values for the data points
    - :param graph_widget: graph widget that will be plotted to (supplied by decorator)

    :return: layout containing the graph (returned by decorator)
    """
    bottom_axis = graph_widget.getAxis('bottom')
    bottom_axis.unit = 's'
    legend_data = list()

    for (player, graph_data), color in zip(data.items(), self.theme['plot']['color_cycler']):
        if player in time_reference:
            time_data = time_reference[player]
            graph_widget.plot(time_data, graph_data, pen=mkPen(color, width=1.5), name=player)
            legend_data.append((color, player))
    return legend_data

def create_legend(self, colors_and_names: Iterable[tuple]) -> QFrame:
    """
    Creates Legend from color / name pairs and returns frame containing it.

    Parameters:
    - :param colors_and_names: Iterable containing color / name pairs : [('#9f9f00', 'Line 1'), 
    ('#0000ff', 'Line 2'), (...), ...]

    :return: frame containing the legend
    """
    frame = create_frame(self, style='plot_legend')
    upper_frame = create_frame(self, style='plot_legend')
    lower_frame = create_frame(self, style='plot_legend')
    frame_layout = QVBoxLayout()
    upper_layout = QHBoxLayout()
    lower_layout = QHBoxLayout()
    margin = self.theme['defaults']['margin']
    frame_layout.setContentsMargins(0, 0, 0, 0)
    frame_layout.setSpacing(margin)
    upper_layout.setContentsMargins(0, 0, 0, 0)
    upper_layout.setSpacing(2 * margin)
    lower_layout.setContentsMargins(0, 0, 0, 0)
    lower_layout.setSpacing(2 * margin)
    second_row = False
    for num, (color, name) in enumerate(colors_and_names, 1):
        legend_item = create_legend_item(self, color, name)
        if num <= 5:
            upper_layout.addWidget(legend_item)
        else:
            second_row = True
            lower_layout.addWidget(legend_item)
    upper_frame.setLayout(upper_layout)
    frame_layout.addWidget(upper_frame, alignment=ACENTER)
    if second_row:
        lower_frame.setLayout(lower_layout)
        frame_layout.addWidget(lower_frame, alignment=ACENTER)
    frame.setLayout(frame_layout)
    return frame

def create_legend_item(self, color: str, name: str) -> QFrame:
    """
    Creates a colored patch next to a label inside a frame

    Parameters:
    - :param color: patch color
    - :param name: text of the label

    :return: frame containing the legend item
    """
    frame = create_frame(self, style='plot_legend')
    layout = QHBoxLayout()
    layout.setContentsMargins(0, 0, 0, 0)
    layout.setSpacing(self.theme['defaults']['margin'])
    colored_patch = QLabel()
    colored_patch.setStyleSheet(get_style(self, 'plot_legend', {'background-color': color}))
    patch_height = self.theme['app']['frame_thickness']
    colored_patch.setFixedSize(2*patch_height, patch_height)
    layout.addWidget(colored_patch, alignment=AVCENTER)
    label = create_label(self, name, 'label', style_override={'font': self.theme['plot_legend']['font']})
    layout.addWidget(label)
    frame.setLayout(layout)
    return frame

def create_overview_table(self) -> QTableView:
    """
    Creates the overview table and returns it.

    :return: Overview Table
    """
    table_data = self.parser1.active_combat.table
    table_cell_data = tuple(tuple(line[2:]) for line in table_data)
    table_index = tuple(line[0] + line[1] for line in table_data)
    model = OverviewTableModel(table_cell_data, TABLE_HEADER, table_index, 
            self.theme_font('table_header'), self.theme_font('table'))
    sort = SortingProxy()
    sort.setSourceModel(model)
    table = QTableView(self.widgets.overview_tab_frames[0])
    table.setModel(sort)
    style_table(self, table)
    for col in range(len(model._header)):
        table.horizontalHeader().resizeSection(col, table.horizontalHeader().sectionSize(col) + 5)
    return table



"""Backend inteface to the OSCR web server"""

import gzip
import tempfile
import os

import OSCR_django_client
from OSCR.utilities import logline_to_str
from OSCR_django_client.api import CombatlogApi, LadderApi, LadderEntriesApi
from PySide6.QtWidgets import QMessageBox

from .datamodels import LeagueTableModel, SortingProxy
from .textedit import format_datetime_str
from .style import theme_font

LADDER_HEADER = ('Name', 'Handle', 'DPS', 'Total Damage', 'Deaths', 'Combat Time', 'Date', 'Max One Hit',
        'Debuff')

def establish_league_connection(self, fetch_maps: bool = False):
    """
    Connects to the league server if not already connected.

    Parameters:
    - :param fetch_ladder: fetches available maps and updates map selector if true
    """
    if self.league_api is None:
        self.league_api = OSCRClient()
    if fetch_maps and isinstance(self.league_api, OSCRClient):
        ladders = self.league_api.ladders()
        if ladders is not None:
            for ladder in ladders.results:
                solo = '[Solo] ' if ladder.is_solo else ''
                key = f'{solo}{ladder.metric} - {ladder.name} ({ladder.difficulty})'
                self.league_api.ladder_dict[key] = ladder
                self.widgets.ladder_map.addItem(key)

def update_ladder_index(self, index):
    """Open Combat Log Dialog Box"""

    selected_map = self.widgets.ladder_map.currentText()
    if not selected_map in self.league_api.ladder_dict:
        return
    
    selected_ladder = self.league_api.ladder_dict[selected_map]
    ladder_data = self.league_api.ladder_entries(selected_ladder.id)
    table_index = list()
    table_data = list()

    for rank, entry in enumerate(ladder_data.results, 1):
        table_index.append(rank)
        row = entry.data
        table_data.append((row['name'], row['handle'], row['DPS'], row['total_damage'], row['deaths'], 
                row['combat_time'], format_datetime_str(entry.var_date), row['max_one_hit'], row['debuff']))
        
    model = LeagueTableModel(table_data, LADDER_HEADER, table_index, theme_font(self, 'table_header'),
            theme_font(self, 'table'))
    sorting_proxy = SortingProxy()
    sorting_proxy.setSourceModel(model)
    table = self.widgets.ladder_table
    table.setModel(sorting_proxy)
    table.resizeColumnsToContents()
    table_header = table.horizontalHeader()
    for col in range(len(model._header)):
        table_header.resizeSection(col, table_header.sectionSize(col) + 5)


def download_and_view_combat(self, idx):
    """
    Download a combat log and view its contents in the overview / analysis pages.
    """


def upload_callback(self):
    """
    Helper function to grab the current combat and upload it to the backend.
    """

    establish_league_connection(self)

    if (
        self.parser1.active_combat is None
        or self.parser1.active_combat.log_data is None
    ):
        raise Exception("No data to upload")

    with tempfile.NamedTemporaryFile(dir=self.app_dir, delete=False) as file:
        data = gzip.compress(
            "".join(
                [logline_to_str(line) for line in self.parser1.active_combat.log_data]
            ).encode()
        )
        file.write(data)
        file.flush()
    self.league_api.upload(file.name)
    os.remove(file.name)


class OSCRClient:
    def __init__(self, address=None):
        """Initialize an instance of the OSCR backlend client"""

        # TODO: This is a test domain and not for production.
        if not address:
            self.address = "https://oscr-server.vercel.app"
            # self.address = "http://127.0.0.1:8000"

        self.api_client = OSCR_django_client.api_client.ApiClient()
        self.api_client.configuration.host = self.address
        self.api_combatlog = CombatlogApi(api_client=self.api_client)
        self.api_ladder = LadderApi(api_client=self.api_client)
        self.api_ladder_entries = LadderEntriesApi(api_client=self.api_client)
        self.ladder_dict: dict = dict()

    def upload(self, filename):
        """Upload a combat log located at path for analysis"""

        reply = QMessageBox()
        reply.setWindowTitle("Open Source Combatlog Reader")

        try:
            res = self.api_combatlog.combatlog_upload(file=filename)
            lines = []
            for entry in res:
                lines.append(entry.detail)
            reply.setText("\n".join(lines))
        except OSCR_django_client.exceptions.ServiceException as e:
            reply.setText(str(e))

        reply.exec()

    def download(self, id):
        """Download a combat log"""
        try:
            return self.api_combatlog.combatlog_download(id=id)
        except OSCR_django_client.exceptions.ServiceException as e:
            reply = QMessageBox()
            reply.setWindowTitle("Open Source Combatlog Reader")
            reply.setText(str(e))
            reply.exec()

        return None

    def ladders(self):
        """Fetch the list of ladders"""
        try:
            return self.api_ladder.ladder_list()
        except OSCR_django_client.exceptions.ServiceException as e:
            reply = QMessageBox()
            reply.setWindowTitle("Open Source Combatlog Reader")
            reply.setText(str(e))
            reply.exec()

        return None

    def ladder_entries(self, id, page=1):
        """Fetch the nth page of ladder entries"""
        try:
            return self.api_ladder_entries.ladder_entries_list(
                ladder=str(id),
                page=page,
                ordering="-data__DPS",
                page_size=200,
            )
        except OSCR_django_client.exceptions.ServiceException as e:
            reply = QMessageBox()
            reply.setWindowTitle("Open Source Combatlog Reader")
            reply.setText(str(e))
            reply.exec()

        return None



from multiprocessing import Pipe, Process
import os

from PySide6.QtCore import QThread, Signal, Qt

from OSCR import OSCR

from .datamodels import DamageTreeModel, HealTreeModel
from .displayer import create_overview
from .widgetbuilder import show_warning


class CustomThread(QThread):
    """
    Subclass of QThread able to execute an arbitrary function in a seperate thread.
    """
    result = Signal(tuple)

    def __init__(self, parent, func) -> None:
        self._func = func
        super().__init__(parent)

    def run(self):
        r = self._func()
        self.result.emit((r,))

def init_parser(self):
    """
    Initializes Parser.
    """
    self.parser1 = OSCR()
    # self.parser2 = OSCR()

def analyze_log_callback(self, combat_id=None, path=None, parser_num: int = 1):
    """
    Wrapper function for retrieving and showing data. Callback of "Analyse" and "Refresh" button.
    
    Parameters:
    - :param combat_id: id of older combat (0 -> latest combat in the file; len(...) - 1 -> oldest combat)
    - :param path: path to combat log file
    """
    if parser_num == 1:
        parser: OSCR = self.parser1
    elif parser_num == 2:
        parser: OSCR = self.parser2
    else:
        return
    
    # initial run / click on the Analyze buttonQGuiApplication
    if combat_id is None:
        if not path or not os.path.isfile(path):
            show_warning(self, 'Invalid Logfile', 'The Logfile you are trying to open does not exist.')
            return
        if path != self.settings.value('log_path'):
            self.settings.setValue('log_path', path)

        get_data(self, combat=None, path=path)
        self.current_combats.clear()
        self.current_combats.addItems(parser.analyzed_combats)
        self.current_combats.setCurrentRow(0)
        self.current_combat_id = 0
        self.current_combat_path = path

        analysis_thread = CustomThread(self.window, lambda: parser.full_combat_analysis(0))
        analysis_thread.result.connect(lambda result: analysis_data_slot(self, result))
        analysis_thread.start(QThread.Priority.IdlePriority)

    # subsequent run / click on older combat
    elif isinstance(combat_id, int) and combat_id != self.current_combat_id:
        if combat_id == -1: return
        get_data(self, combat_id)
        self.current_combat_id = combat_id
        analysis_thread = CustomThread(self.window, lambda: parser.full_combat_analysis(combat_id))
        analysis_thread.result.connect(lambda result: analysis_data_slot(self, result))
        analysis_thread.start(QThread.Priority.IdlePriority)

    create_overview(self)

    # reset tabber
    self.widgets.main_tabber.setCurrentIndex(0)
    self.widgets.overview_tabber.setCurrentIndex(0)

def copy_summary_callback(self):
    """
    Callback to set the combat summary of the active combat to the user's clippboard
    """

    if not self.parser1.active_combat:
        return

    parts = [
        "OSCR",
        f"{self.parser1.active_combat.map}",
        f"{self.parser1.active_combat.difficulty}",
        "DPS",
    ]
    players = sorted(
        self.parser1.active_combat.player_dict.items(),
        reverse=True,
        key=lambda player: player[1].DPS,
    )
    for player in players:
        parts.append(f"{player[1].handle} {player[1].DPS:,.0f}")
    summary = " | ".join(parts)

    self.app.clipboard().setText(summary)


def get_data(self, combat: int | None = None, path: str | None = None):
    """Interface between OSCRUI and OSCR. 
    Uses OSCR class to analyze log at path"""

    # new log file
    if combat is None:
        self.parser1.log_path = path
        try:
            self.parser1.analyze_log_file()
        except FileExistsError:
            # TODO show annoying message prompting to split the logfile
            self.parser1.analyze_massive_log_file()
        self.parser1.shallow_combat_analysis(0)
        
    # same log file, old combat
    else:
        self.parser1.shallow_combat_analysis(combat)

def analysis_data_slot(self, item_tuple: tuple):
    """
    Inserts the data retrieved from the parser into the respective tables

    Parameters:
    - :param item_tuple: tuple containing only the root item of the data model
    """
    populate_analysis(self, *item_tuple)
    self.widgets.main_menu_buttons[1].setDisabled(False)

def populate_analysis(self, root_items: tuple):
    """
    Populates the Analysis' treeview table.
    """
    damage_out_item, damage_in_item, heal_out_item, heal_in_item = root_items

    damage_out_table = self.widgets.analysis_table_dout
    damage_out_model = DamageTreeModel(damage_out_item, self.theme_font('tree_table_header'),
            self.theme_font('tree_table'),
            self.theme_font('', self.theme['tree_table']['::item']['font']))
    damage_out_table.setModel(damage_out_model)
    damage_out_table.expand(damage_out_model.index(0, 0, 
            damage_out_model.createIndex(0, 0, damage_out_model._root)))
    damage_out_table.sortByColumn(1, Qt.SortOrder.AscendingOrder)

    damage_in_table = self.widgets.analysis_table_dtaken
    damage_in_model = DamageTreeModel(damage_in_item, self.theme_font('tree_table_header'),
            self.theme_font('tree_table'),
            self.theme_font('', self.theme['tree_table']['::item']['font']))
    damage_in_table.setModel(damage_in_model)
    damage_in_table.expand(damage_in_model.index(0, 0, 
            damage_in_model.createIndex(0, 0, damage_in_model._root)))
    damage_in_table.sortByColumn(1, Qt.SortOrder.AscendingOrder)

    heal_out_table = self.widgets.analysis_table_hout
    heal_out_model = HealTreeModel(heal_out_item, self.theme_font('tree_table_header'),
            self.theme_font('tree_table'),
            self.theme_font('', self.theme['tree_table']['::item']['font']))
    heal_out_table.setModel(heal_out_model)
    heal_out_table.expand(heal_out_model.index(0, 0, 
            damage_in_model.createIndex(0, 0, heal_out_model._root)))
    heal_out_table.sortByColumn(1, Qt.SortOrder.AscendingOrder)

    heal_in_table = self.widgets.analysis_table_hin
    heal_in_model = HealTreeModel(heal_in_item, self.theme_font('tree_table_header'),
            self.theme_font('tree_table'),
            self.theme_font('', self.theme['tree_table']['::item']['font']))
    heal_in_table.setModel(heal_in_model)
    heal_in_table.expand(heal_in_model.index(0, 0, 
            damage_in_model.createIndex(0, 0, heal_in_model._root)))
    heal_in_table.sortByColumn(1, Qt.SortOrder.AscendingOrder)
    
    update_shown_columns_dmg(self)
    update_shown_columns_heal(self)

def update_shown_columns_dmg(self):
    """
    Hides / shows columns of the dmg analysis tables.
    """
    dout_table = self.widgets.analysis_table_dout
    dtaken_table = self.widgets.analysis_table_dtaken
    for i in range(self.settings.value('dmg_columns_length', type=int)):
        state = self.settings.value(f'dmg_columns|{i}')
        if state:
            dout_table.showColumn(i+1)
            dtaken_table.showColumn(i+1)
        else:
            dout_table.hideColumn(i+1)
            dtaken_table.hideColumn(i+1)

def update_shown_columns_heal(self):
    """
    Hides / shows columns of the heal analysis tables.
    """
    hout_table = self.widgets.analysis_table_hout
    hin_table = self.widgets.analysis_table_hin
    for i in range(self.settings.value('heal_columns_length', type=int)):
        state = self.settings.value(f'heal_columns|{i}')
        if state:
            hout_table.showColumn(i+1)
            hin_table.showColumn(i+1)
        else:
            hout_table.hideColumn(i+1)
            hin_table.hideColumn(i+1)
        
def resize_tree_table(tree):
    """
    Resizes the columns of the given tree table to fit its contents.

    Parameters:
    - :param tree: QTreeView -> tree to be resized
    """
    for col in range(tree.header().count()):
        width = max(tree.sizeHintForColumn(col), tree.header().sectionSizeHint(col)) + 5
        tree.header().resizeSection(col, width)



from typing import Iterable

from PySide6.QtCore import Qt, QAbstractTableModel, QSortFilterProxyModel, QAbstractItemModel, QModelIndex
from PySide6.QtGui import QFont
from OSCR import TreeItem

from .widgetbuilder import AVCENTER, ARIGHT, ACENTER, ALEFT

class TableModel(QAbstractTableModel):
    def __init__(self, data, header: Iterable, index: Iterable, header_font: QFont, cell_font: QFont):
        """
        Creates table model from supplied data.

        Parameters:
        - :param data: data to be displayed without index or header; two-dimensional iterable
        - :param header: column headings
        - :param index: row index
        - :param header_font: font to style the column headings with
        - :param cell_font: font to style the cells with
        """
        super().__init__()
        self._data = data
        self._header = tuple(header)
        self._index = list(index)
        self._header_font = header_font
        self._cell_font = cell_font

    def rowCount(self, index):
        return len(self._data)

    def columnCount(self, index):
        try:
            return len(self._data[0]) # all columns must have the same length
        except IndexError:
            return 0

    def headerData(self, section, orientation, role):
        # section is the index of the column/row.
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                return self._header[section]

            if orientation == Qt.Orientation.Vertical:
                return self._index[section]

        if role == Qt.ItemDataRole.FontRole:
            return self._header_font

        if role == Qt.ItemDataRole.TextAlignmentRole:
            if orientation == Qt.Orientation.Horizontal:
                return ACENTER

            if orientation == Qt.Orientation.Vertical:
                return AVCENTER + ARIGHT
            
class OverviewTableModel(TableModel):
    """
    Model for overview table
    """
    def data(self, index, role):
        if role == Qt.ItemDataRole.DisplayRole:
            current_col = index.column()
            cell = self._data[index.row()][current_col]
            column = index.column()
            if column == 0:
                return f'{cell:.1f}s'
            elif column in (1, 2, 7, 10, 13, 14, 15):
                return f'{cell:,.2f}'
            elif column in (3, 4, 5, 6, 8, 11, 12):
                return f'{cell:,.2f}%'
            elif column in (9, 16, 17, 18, 19, 20, 21, 22):
                return str(cell)
            return cell
        
        if role == Qt.ItemDataRole.FontRole:
            return self._cell_font

        if role == Qt.ItemDataRole.TextAlignmentRole:
            return AVCENTER + ARIGHT
        
class LeagueTableModel(TableModel):
    """
    Model for league table
    """
    def data(self, index, role):
        if role == Qt.ItemDataRole.DisplayRole:
            current_col = index.column()
            cell = self._data[index.row()][current_col]
            column = index.column()
            if column == 5:
                return f'{cell:.1f}s'
            elif column in (2, 3, 7):
                return f'{cell:,.2f}'
            elif column == 8:
                return f'{cell:,.2f}%'
            elif column == 4:
                return str(cell)
            return cell
        
        if role == Qt.ItemDataRole.FontRole:
            return self._cell_font

        if role == Qt.ItemDataRole.TextAlignmentRole:
            if index.column() == 1:
                return AVCENTER + ALEFT
            return AVCENTER + ARIGHT
    
    def headerData(self, section, orientation, role):
        if role == Qt.ItemDataRole.FontRole and orientation == Qt.Orientation.Vertical:
            return self._cell_font
        return super().headerData(section, orientation, role)

class SortingProxy(QSortFilterProxyModel):
    def __init__(self):
        super().__init__()

    def lessThan(self, left, right):
        l = self.sourceModel()._data[left.row()][left.column()]
        r = self.sourceModel()._data[right.row()][right.column()]
        return l > r # inverted operator to make descending sort come up first
    
class TreeModel(QAbstractItemModel):
    """
    Data model for the analysis table
    """
    def __init__(self, root_item, header_font: QFont, name_font: QFont, cell_font: QFont):
        """
        Initializes Tree Model with data in root item.

        Parameters:
        - :param root_item: item supporting the following operations:
            - function "get_child(row: int)" returning the n-th child of the item; None if not exists
            - function "get_data(column: int)" returning the data of the item at the given column; None if
            not exists
            - function "append_child(item)" adds the given item as child to the item 
            - property "parent" containing the parent item of the item; None for the root item
            - property "row" containing the row number which the item is stored at in its parent item
            - property "child_count" containing the number of children the item has
            - property "column_count" containing the number of columns the items data row

        The item may already contain childen.
        - :param header_font: font used for the header
        - :param name_font: font used for the first column
        - :param cell_font: font used for the second to last column
        """
        super().__init__()
        self._root = root_item
        self._root_index = self.createIndex(0, 0, self._root)
        self._header_font = header_font
        self._name_font = name_font
        self._cell_font = cell_font

    def sort(self, column: int, order: Qt.SortOrder):
        if order == Qt.SortOrder.AscendingOrder:
            descending = True
        else:
            descending = False
        self.layoutAboutToBeChanged.emit()
        self.recursive_sort(self._root._children[0], column, descending)
        self.recursive_sort(self._root._children[1], column, descending)
        self.layoutChanged.emit()
    
    def recursive_sort(self, item: TreeItem, column: int, desc):
        if item.child_count > 0:
            for child in item._children:
                self.recursive_sort(child, column, desc)
            item._children.sort(key=lambda row: row.get_data(column), reverse=desc)

    def index(self, row: int, column: int, parent: QModelIndex) -> QModelIndex | None:
        if not self.hasIndex(row, column, parent):
            return QModelIndex()
        if not parent.isValid():
            p = self._root
        else:
            p = parent.internalPointer()
        c = p.get_child(row)
        if not c is None:
            return self.createIndex(row, column, c)
        return QModelIndex()
    
    def parent(self, index: QModelIndex) -> QModelIndex | None:
        if not index.isValid():
            return QModelIndex()
        child = index.internalPointer()
        parent = child.parent
        if parent is None:
            return QModelIndex()
        return self.createIndex(parent.row, 0, parent)
    
    def rowCount(self, parent: QModelIndex) -> int:
        if not parent.isValid():
            parent = self._root
        else:
            parent = parent.internalPointer()
        return parent.child_count
    
    def columnCount(self, parent: QModelIndex) -> int:
        if parent.isValid():
            return parent.internalPointer().column_count
        return self._root.column_count
    
    def flags(self, index: QModelIndex) -> Qt.ItemFlag:
        if not index.isValid():
            return Qt.ItemFlag.NoItemFlags
        return super().flags(index)
    
    def headerData(self, section, orientation, role) -> str:
        if role == Qt.ItemDataRole.DisplayRole:
            return self._root.data[section]
        elif role == Qt.ItemDataRole.FontRole:
            return self._header_font
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            return ACENTER
        return None


class DamageTreeModel(TreeModel):
    """
    Tree Model subclass for the damage tables
    """
    def data(self, index: QModelIndex, role: int) -> str:
        if not index.isValid():
            return ''
        column = index.column()
        if role == Qt.ItemDataRole.DisplayRole:
            data = index.internalPointer().get_data(column)
            if data == '':
                return ''
            if column == 0:
                if isinstance(data, tuple):
                    return ''.join(data)
                return data
            elif column in (3, 5, 6, 7):
                return f'{data * 100:,.2f}%'
            elif column in (1, 2, 4, 13, 14, 15, 16, 17, 18):
                return f'{data:,.2f}'
            elif column in (8, 9, 10, 11, 12, 20, 21):
                return f'{data:,.0f}'
            elif column == 19:
                return f'{data}s'
        elif role == Qt.ItemDataRole.FontRole:
            if column == 0:
                return self._name_font
            return self._cell_font
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            if column != 0:
                return AVCENTER + ARIGHT
            else:
                return AVCENTER + ALEFT
        elif role == -13:
            return index.internalPointer().get_data(column)
        return None

class HealTreeModel(TreeModel):
    """
    Tree Model subclass for the heal tables
    """
    def data(self, index: QModelIndex, role: int) -> str:
        if not index.isValid():
            return ''
        column = index.column()
        if role == Qt.ItemDataRole.DisplayRole:
            data = index.internalPointer().get_data(column)
            if data == '':
                return ''
            if column == 0:
                if isinstance(data, tuple):
                    return ''.join(data)
                return data
            elif column == 8:
                return f'{data * 100:,.2f}%'
            elif column in (1, 2, 3, 4, 5, 6, 7, 17, 18):
                return f'{data:,.2f}'
            elif column in (9, 10, 12, 13):
                return f'{data:,.0f}'
            elif column == 11:
                return f'{data}s'
        elif role == Qt.ItemDataRole.FontRole:
            if column == 0:
                return self._name_font
            return self._cell_font
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            if column != 0:
                return AVCENTER + ARIGHT
            else:
                return AVCENTER + ALEFT
        elif role == -13:
            return index.internalPointer().get_data(column)
        return None
    


from signal import signal, SIGINT, SIG_DFL
from multiprocessing import Lock
import os

from PySide6.QtWidgets import QApplication, QWidget, QLineEdit, QFrame, QListWidget, QTabWidget, QTableView
from PySide6.QtWidgets import QVBoxLayout, QHBoxLayout, QGridLayout
from PySide6.QtCore import QSize, QSettings

from OSCR import TREE_HEADER, HEAL_TREE_HEADER
from OSCR.utilities import logline_to_str

from .iofunctions import load_icon_series, get_asset_path, load_icon
from .textedit import format_path
from .widgets import BannerLabel, FlipButton, WidgetStorage, AnalysisPlot
from .widgetbuilder import SMAXMAX, SMAXMIN, SMINMAX, SMINMIN, ALEFT, ARIGHT, ATOP, ACENTER, AHCENTER
from .leagueconnector import OSCRClient

# only for developing; allows to terminate the qt event loop with keyboard interrupt
signal(SIGINT, SIG_DFL)

class OSCRUI():

    from .datafunctions import init_parser, copy_summary_callback
    from .datafunctions import analyze_log_callback, update_shown_columns_dmg, update_shown_columns_heal
    from .displayer import create_legend_item
    from .iofunctions import browse_path
    from .style import get_style_class, create_style_sheet, theme_font, get_style
    from .widgetbuilder import create_frame, create_label, create_button_series, create_icon_button
    from .widgetbuilder import create_analysis_table, create_button, create_combo_box, style_table
    from .leagueconnector import upload_callback, update_ladder_index, establish_league_connection

    app_dir = None

    config = {} # see main.py for contents

    settings: QSettings # see main.py for defaults

    # stores widgets that need to be accessed from outside their creating function
    widgets: WidgetStorage

    league_api: OSCRClient

    def __init__(self, version, theme, args, path, config) -> None:
        """
        Creates new Instance of OSCR.

        Parameters:
        - :param version: version of the app
        - :param theme: dict -> default theme
        - :param args: command line arguments
        - :param path: absolute path to main.py file
        - :param config: app configuration (!= settings these are not changed by the user)
        """
        self.version = version
        self.theme = theme
        self.args = args
        self.app_dir = path
        self.config = config
        self.widgets = WidgetStorage()
        self.league_api = None
        self.init_settings()
        self.app, self.window = self.create_main_window()
        self.init_config()
        self.init_parser()
        self.cache_assets()
        self.setup_main_layout()
        self.window.show()

    def run(self) -> int:
        """
        Runs the event loop.

        :return: exit code of event loop
        """
        return self.app.exec()
    
    def cache_assets(self):
        """
        Caches assets like icon images
        """
        icons = {
            'oscr': 'oscr_icon_small.png',
            'expand-left': 'expand-left.svg',
            'collapse-left': 'collapse-left.svg',
            'expand-right': 'expand-right.svg',
            'collapse-right': 'collapse-right.svg',
            'save': 'save-cw.svg',
            'truncate': 'truncate-cw.svg',
            'refresh': 'refresh-cw.svg'
        }
        self.icons = load_icon_series(icons, self.app_dir)

    def init_settings(self):
        """
        Prepares settings. Loads stored settings. Saves current settings for next startup.
        """
        settings_path = os.path.abspath(f'{self.app_dir}/{self.config["settings_path"]}')
        self.settings = QSettings(settings_path, QSettings.Format.IniFormat)
        for setting, value in self.config['default_settings'].items():
            if self.settings.value(setting, None) is None:
                self.settings.setValue(setting, value)
        if not self.settings.value('log_path', ''):
            self.settings.setValue('log_path', format_path(self.app_dir))
        
    def init_config(self):
        """
        Prepares config.
        """
        self.current_combat_id = -1
        self.current_combat_path = ''

    @property
    def sidebar_item_width(self) -> int:
        """
        Width of the sidebar.
        """
        return int(self.theme['s.c']['sidebar_item_width'] * self.window.width())

    def main_window_close_callback(self, event):
        """
        Executed when application is closed.
        """
        window_geometry = self.window.saveGeometry()
        self.settings.setValue('geometry', window_geometry)
        event.accept()

    def main_window_resize_callback(self, event):
        """
        Executed when application is resized.
        """
        self.entry.setFixedWidth(self.sidebar_item_width)
        self.current_combats.setFixedWidth(self.sidebar_item_width)
        event.accept()
    
    # -------------------------------------------------------------------------------------------------------
    # GUI functions below
    # -------------------------------------------------------------------------------------------------------

    def create_main_window(self, argv=[]) -> tuple[QApplication, QWidget]:
        """
        Creates and initializes main window.

        :return: QApplication, QWidget
        """
        app = QApplication(argv)
        app.setStyleSheet(self.create_style_sheet(self.theme['app']['style']))
        window = QWidget()
        window.setWindowIcon(load_icon('oscr_icon_small.png', self.app_dir))
        window.setWindowTitle('Open Source Combatlog Reader')
        window.setMinimumSize(
                int(self.config['minimum_window_width']),
                int(self.config['minimum_window_height']))
        if self.settings.value('geometry'):
            window.restoreGeometry(self.settings.value('geometry'))        
        window.closeEvent = self.main_window_close_callback
        window.resizeEvent = self.main_window_resize_callback
        return app, window

    def setup_main_layout(self):
        """
        Sets up the main layout of the app.
        """
        layout, main_frame = self.create_master_layout(self.window)
        self.window.setLayout(layout)

        content_layout = QGridLayout()
        content_layout.setContentsMargins(0, 0, 0, 0)
        content_layout.setSpacing(0)

        left = self.create_frame(main_frame, 'medium_frame')
        left.setSizePolicy(SMAXMIN)
        content_layout.addWidget(left, 0, 0)

        right = self.create_frame(main_frame, 'frame', 
                {'border-left-style': 'solid', 'border-left-width': '@sep', 'border-left-color': '@oscr' })
        right.setSizePolicy(SMINMIN)
        right.hide()
        content_layout.addWidget(right, 0, 4)

        icon_size = self.theme['s.c']['button_icon_size']
        left_flip_config = {
            'icon_r': self.icons['collapse-left'], 'func_r': left.hide,
            'icon_l': self.icons['expand-left'], 'func_l': left.show
        }
        right_flip_config = {
            'icon_r': self.icons['expand-right'], 'func_r': right.show,
            'icon_l': self.icons['collapse-right'], 'func_l': right.hide
        }
        for col, config in ((1, left_flip_config), (3, right_flip_config)):
            flip_button = FlipButton('', '', main_frame)
            flip_button.configure(config)
            flip_button.setIconSize(QSize(icon_size, icon_size))
            flip_button.setStyleSheet(self.get_style_class('QPushButton', 'small_button'))
            flip_button.setSizePolicy(SMAXMAX)
            content_layout.addWidget(flip_button, 0, col, alignment=ATOP)

        center = self.create_frame(main_frame, 'frame')
        center.setSizePolicy(SMINMIN)
        content_layout.addWidget(center, 0, 2)

        main_frame.setLayout(content_layout)
        self.setup_left_sidebar(left)
        self.setup_main_tabber(center)
        self.setup_overview_frame()
        self.setup_analysis_frame()
        self.setup_league_standings_frame()
        self.setup_settings_frame()

    def setup_left_sidebar(self, frame:QFrame):
        """
        Sets up the sidebar used to select parses and combats

        Parameters:
        - :param frame: QFrame -> parent frame of the sidebar
        """
        m = self.theme['defaults']['margin']
        left_layout = QVBoxLayout()
        left_layout.setContentsMargins(m, m, m, m)
        left_layout.setSpacing(0)
        left_layout.setAlignment(ATOP)
        left_button_layout = QHBoxLayout()
        left_button_layout.setContentsMargins(m, m, m, m)
        left_button_layout.setSpacing(0)
        left_button_layout.setAlignment(ATOP)

        head = self.create_label('STO Combatlog:', 'label', frame)
        left_layout.addWidget(head, alignment=ALEFT)

        self.entry = QLineEdit(self.settings.value('log_path', ''), frame)
        self.entry.setStyleSheet(self.get_style_class('QLineEdit', 'entry'))
        self.entry.setFixedWidth(self.sidebar_item_width)
        left_layout.addWidget(self.entry)
        
        button_frame = self.create_frame(frame, 'medium_frame')
        button_frame.setSizePolicy(SMINMAX)
        entry_button_config = {
            'default': {'margin-bottom': '@isp'},
            'Browse ...': {'callback': lambda: self.browse_log(self.entry), 'align': ALEFT},
            'Analyze': {'callback': lambda: self.analyze_log_callback(path=self.entry.text()), 
                    'align': ARIGHT}
        }
        entry_buttons = self.create_button_series(button_frame, entry_button_config, 'button')
        button_frame.setLayout(entry_buttons)
        left_layout.addWidget(button_frame)

        save_button = self.create_icon_button(self.icons['save'], 'icon_button', frame)
        left_button_layout.addWidget(save_button, alignment=ALEFT)
        truncate_button = self.create_icon_button(self.icons['truncate'])
        left_button_layout.addWidget(truncate_button, alignment=ARIGHT)
        left_layout.addLayout(left_button_layout)
        refresh_button = self.create_icon_button(self.icons['refresh'], 'icon_button', frame)
        left_layout.addWidget(refresh_button, alignment=ARIGHT)

        background_frame = self.create_frame(frame, 'light_frame', 
                {'border-radius': self.theme['listbox']['border-radius']})
        background_frame.setSizePolicy(SMAXMIN)
        background_layout = QVBoxLayout()
        background_layout.setContentsMargins(0, 0, 0, 0)
        background_frame.setLayout(background_layout)
        self.current_combats = QListWidget(background_frame)
        self.current_combats.setStyleSheet(self.get_style_class('QListWidget', 'listbox'))
        self.current_combats.setFont(self.theme_font('listbox'))
        self.current_combats.setSizePolicy(SMAXMIN)
        self.current_combats.setFixedWidth(self.sidebar_item_width)
        background_layout.addWidget(self.current_combats)
        left_layout.addWidget(background_frame, stretch=1)
        
        refresh_button.clicked.connect(lambda: self.analyze_log_callback(self.current_combats.currentRow()))
        save_button.clicked.connect(lambda: self.save_log(self.entry))
        truncate_button.clicked.connect(lambda: self.truncate_log(self.entry))

        frame.setLayout(left_layout)

    def setup_main_tabber(self, frame:QFrame):
        """
        Sets up the tabber switching between Overview, Analysis, League and Settings.

        Parameters:
        - :param frame: QFrame -> parent frame of the sidebar
        """
        o_frame = self.create_frame(None, 'frame')
        a_frame = self.create_frame(None, 'frame')
        l_frame = self.create_frame(None, 'frame')
        s_frame = self.create_frame(None, 'frame')

        main_tabber = QTabWidget(frame)
        main_tabber.setStyleSheet(self.get_style_class('QTabWidget', 'tabber'))
        main_tabber.tabBar().setStyleSheet(self.get_style_class( 'QTabBar', 'tabber_tab'))
        main_tabber.setSizePolicy(SMINMIN)
        main_tabber.addTab(o_frame, 'O')
        main_tabber.addTab(a_frame, 'A')
        main_tabber.addTab(l_frame, 'L')
        main_tabber.addTab(s_frame, 'S')

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(main_tabber)
        frame.setLayout(layout)

        self.widgets.main_menu_buttons[0].clicked.connect(lambda: main_tabber.setCurrentIndex(0))
        self.widgets.main_menu_buttons[1].clicked.connect(lambda: main_tabber.setCurrentIndex(1))
        self.widgets.main_menu_buttons[2].clicked.connect(lambda: main_tabber.setCurrentIndex(2))
        self.widgets.main_menu_buttons[2].clicked.connect(lambda: self.establish_league_connection(True))
        self.widgets.main_menu_buttons[3].clicked.connect(lambda: main_tabber.setCurrentIndex(3))
        self.widgets.main_tab_frames.append(o_frame)
        self.widgets.main_tab_frames.append(a_frame)
        self.widgets.main_tab_frames.append(l_frame)
        self.widgets.main_tab_frames.append(s_frame)
        self.widgets.main_tabber = main_tabber

    def setup_overview_frame(self):
        """
        Sets up the frame housing the combatlog overview
        """
        o_frame = self.widgets.main_tab_frames[0]
        bar_frame = self.create_frame(None, 'frame')
        dps_graph_frame = self.create_frame(None, 'frame')
        dmg_graph_frame = self.create_frame(None, 'frame')
        self.widgets.overview_tab_frames.append(bar_frame)
        self.widgets.overview_tab_frames.append(dps_graph_frame)
        self.widgets.overview_tab_frames.append(dmg_graph_frame)
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        switch_frame = self.create_frame(o_frame, 'frame')
        layout.addWidget(switch_frame, alignment=ACENTER)

        o_tabber = QTabWidget(o_frame)
        o_tabber.setStyleSheet(self.get_style_class('QTabWidget', 'tabber'))
        o_tabber.tabBar().setStyleSheet(self.get_style_class('QTabBar', 'tabber_tab'))
        o_tabber.addTab(bar_frame, 'BAR')
        o_tabber.addTab(dps_graph_frame, 'DPS')
        o_tabber.addTab(dmg_graph_frame, 'DMG')
        layout.addWidget(o_tabber)

        switch_style = {
            'default': {'margin-left': '@margin', 'margin-right': '@margin'},
            'DPS Bar': {'callback': lambda: o_tabber.setCurrentIndex(0), 'align':ACENTER},
            'DPS Graph': {'callback': lambda: o_tabber.setCurrentIndex(1), 'align':ACENTER},
            'Damage Graph': {'callback': lambda: o_tabber.setCurrentIndex(2), 'align':ACENTER},
            'Copy Summary': {'callback': self.copy_summary_callback, 'align':ACENTER},
            'Upload Result': {'callback': self.upload_callback, 'align':ACENTER},
        }
        switcher, buttons = self.create_button_series(switch_frame, switch_style, 'button', ret=True)
        switcher.setContentsMargins(0, self.theme['defaults']['margin'], 0, 0)
        self.widgets.overview_menu_buttons = buttons
        switch_frame.setLayout(switcher)
        o_frame.setLayout(layout)
        self.widgets.overview_tabber = o_tabber

    def setup_analysis_frame(self):
        """
        Sets up the frame housing the detailed analysis table and graph
        """
        a_frame = self.widgets.main_tab_frames[1]
        dout_frame = self.create_frame(None, 'frame')
        dtaken_frame = self.create_frame(None, 'frame')
        hout_frame = self.create_frame(None, 'frame')
        hin_frame = self.create_frame(None, 'frame')
        self.widgets.analysis_tab_frames.extend((dout_frame, dtaken_frame, hout_frame, hin_frame))
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        switch_frame = self.create_frame(a_frame, 'frame')
        layout.addWidget(switch_frame, alignment=ACENTER)
        
        a_tabber = QTabWidget(a_frame)
        a_tabber.setStyleSheet(self.get_style_class('QTabWidget', 'tabber'))
        a_tabber.tabBar().setStyleSheet(self.get_style_class('QTabBar', 'tabber_tab'))
        a_tabber.addTab(dout_frame, 'DOUT')
        a_tabber.addTab(dtaken_frame, 'DTAKEN')
        a_tabber.addTab(hout_frame, 'HOUT')
        a_tabber.addTab(hin_frame, 'HIN')
        self.widgets.analysis_tabber = a_tabber
        layout.addWidget(a_tabber)

        switch_style = {
            'default': {'margin-left': '@margin', 'margin-right': '@margin'},
            'Damage Out': {'callback': lambda: a_tabber.setCurrentIndex(0), 'align':ACENTER},
            'Damage Taken': {'callback': lambda: a_tabber.setCurrentIndex(1), 'align':ACENTER},
            'Heals Out': {'callback': lambda: a_tabber.setCurrentIndex(2), 'align':ACENTER},
            'Heals In': {'callback': lambda: a_tabber.setCurrentIndex(3), 'align':ACENTER}
        }
        switcher, buttons = self.create_button_series(switch_frame, switch_style, 'button', ret=True)
        switcher.setContentsMargins(0, self.theme['defaults']['margin'], 0, 0)
        self.widgets.analysis_menu_buttons = buttons
        switch_frame.setLayout(switcher)

        tabs = (
            (dout_frame, 'analysis_table_dout', 'analysis_plot_dout'), 
            (dtaken_frame, 'analysis_table_dtaken', 'analysis_plot_dtaken'),
            (hout_frame, 'analysis_table_hout', 'analysis_plot_hout'),
            (hin_frame, 'analysis_table_hin', 'analysis_plot_hin')
        )
        for tab, table_name, plot_name in tabs:
            tab_layout = QVBoxLayout()
            tab_layout.setContentsMargins(0, 0, 0, 0)
            tab_layout.setSpacing(0)

            # graph
            plot_frame = self.create_frame(tab, 'plot_widget', {'margin-right': 0}, SMINMAX)
            plot_layout = QHBoxLayout()
            plot_layout.setContentsMargins(0, 0, 0, 0)
            plot_layout.setSpacing(self.theme['defaults']['isp'])
            
            plot_bundle_frame = self.create_frame(plot_frame, size_policy=SMINMAX)
            plot_bundle_layout = QVBoxLayout()
            plot_bundle_layout.setContentsMargins(0, 0, 0, 0)
            plot_bundle_layout.setSpacing(0)
            plot_legend_frame = self.create_frame(plot_bundle_frame)
            plot_legend_layout = QHBoxLayout()
            plot_legend_layout.setContentsMargins(0, 0, 0, 0)
            plot_legend_layout.setSpacing(2 * self.theme['defaults']['margin'])
            plot_legend_frame.setLayout(plot_legend_layout)
            plot_widget = AnalysisPlot(self.theme['plot']['color_cycler'], self.theme['defaults']['fg'],
                    self.theme_font('plot_widget'), plot_legend_layout)
            setattr(self.widgets, plot_name, plot_widget)
            plot_widget.setStyleSheet(self.get_style('plot_widget_nullifier'))
            plot_widget.setSizePolicy(SMINMAX)
            plot_bundle_layout.addWidget(plot_widget)
            plot_bundle_layout.addWidget(plot_legend_frame, alignment=AHCENTER)
            plot_bundle_frame.setLayout(plot_bundle_layout)
            plot_layout.addWidget(plot_bundle_frame, stretch=1)

            plot_button_frame = self.create_frame(plot_frame, size_policy=SMAXMIN)
            plot_button_layout = QVBoxLayout()
            plot_button_layout.setContentsMargins(0, 0, 0, 0)
            plot_button_layout.setSpacing(0)
            freeze_button = self.create_button('Freeze Graph', 'toggle_button', plot_button_frame, 
                    style_override={'border-color': '@bg'}, toggle=True)
            freeze_button.clicked.connect(plot_widget.toggle_freeze)
            plot_button_layout.addWidget(freeze_button, alignment=ARIGHT)
            clear_button = self.create_button('Clear Graph', parent=plot_button_frame)
            clear_button.clicked.connect(plot_widget.clear_plot)
            plot_button_layout.addWidget(clear_button, alignment=ARIGHT)
            plot_button_frame.setLayout(plot_button_layout)
            plot_layout.addWidget(plot_button_frame, stretch=0)

            plot_frame.setLayout(plot_layout)  
            tab_layout.addWidget(plot_frame, stretch=3)

            tree = self.create_analysis_table(tab, 'tree_table')
            setattr(self.widgets, table_name, tree)
            tree.clicked.connect(lambda index, pw=plot_widget: self.slot_analysis_graph(index, pw))
            tab_layout.addWidget(tree, stretch=7)
            tab.setLayout(tab_layout)          
        
        a_frame.setLayout(layout)

    def slot_analysis_graph(self, index, plot_widget: AnalysisPlot):
        item = index.internalPointer()
        color = plot_widget.add_bar(item.graph_data)
        if color is None:
            return
        name = item.data[0]
        if isinstance(name, tuple):
            name = ''.join(name)
        legend_item = self.create_legend_item(color, name)
        plot_widget.add_legend_item(legend_item)

    def setup_league_standings_frame(self):
        """
        Sets up the frame housing the detailed analysis table and graph
        """
        l_frame = self.widgets.main_tab_frames[2]
        layout = QVBoxLayout()
        margin = self.theme['defaults']['margin']
        layout.setContentsMargins(0, 0, 0, margin)
        layout.setSpacing(0)

        spacing = self.theme['defaults']['isp']
        control_frame = self.create_frame(l_frame)
        control_frame_layout = QHBoxLayout()
        control_frame_layout.setContentsMargins(margin, margin, margin, margin)
        control_frame_layout.setSpacing(spacing)

        map_selector = self.create_combo_box(control_frame)
        map_selector.addItem("Select a League")
        self.widgets.ladder_map = map_selector
        map_selector.currentIndexChanged.connect(lambda new_index: self.update_ladder_index(new_index))
        control_frame_layout.addWidget(map_selector)
        control_frame.setLayout(control_frame_layout)

        ladder_table = QTableView(l_frame)
        self.style_table(ladder_table)
        self.widgets.ladder_table = ladder_table

        layout.addWidget(control_frame, alignment=AHCENTER)
        layout.addWidget(ladder_table)
        l_frame.setLayout(layout)

    def create_master_layout(self, parent) -> tuple[QVBoxLayout, QFrame]:
        """
        Creates and returns the master layout for an OSCR window.

        Parameters:
        - :param parent: parent to the layout, usually a window

        :return: populated QVBoxlayout and content frame QFrame
        """
        layout =  QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        bg_frame = self.create_frame(parent, 'frame', {'background': '@oscr'})
        bg_frame.setSizePolicy(SMINMIN)
        layout.addWidget(bg_frame)

        main_layout = QVBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setSpacing(0)
        lbl = BannerLabel(get_asset_path('oscrbanner-slim-dark-label.png', self.app_dir), bg_frame)

        main_layout.addWidget(lbl)

        menu_frame = self.create_frame(bg_frame, 'frame', {'background':'@oscr'})
        menu_frame.setSizePolicy(SMAXMAX)
        menu_frame.setContentsMargins(0, 0, 0, 0)
        main_layout.addWidget(menu_frame)
        menu_button_style = {
            'Overview': {'style':{'margin-left': '@isp'}},
            'Analysis': {},
            'League Standings': {},
            'Settings': {},
        }
        bt_lay, buttons = self.create_button_series(menu_frame, menu_button_style, 
                style='menu_button', seperator='•', ret=True)
        menu_frame.setLayout(bt_lay)
        self.widgets.main_menu_buttons = buttons

        w = self.theme['app']['frame_thickness']
        main_frame = self.create_frame(bg_frame, 'frame', {'margin':(0, w, w, w)})
        main_frame.setSizePolicy(SMINMIN)
        main_layout.addWidget(main_frame)
        bg_frame.setLayout(main_layout)

        return layout, main_frame

    def setup_settings_frame(self):
        """
        Populates the settings frame.
        """
        settings_frame = self.widgets.main_tab_frames[3]
        settings_layout = QHBoxLayout()
        settings_layout.setContentsMargins(0, 0, 0, 0)
        settings_layout.setSpacing(0)

        col_1_frame = self.create_frame(settings_frame)
        col_1_frame.setSizePolicy(SMINMAX)
        settings_layout.addWidget(col_1_frame, alignment=ATOP, stretch=1)
        col_2_frame = self.create_frame(settings_frame)
        col_2_frame.setSizePolicy(SMINMAX)
        settings_layout.addWidget(col_2_frame, alignment=ATOP, stretch=1)
        col_3_frame = self.create_frame(settings_frame)
        col_3_frame.setSizePolicy(SMINMAX)
        settings_layout.addWidget(col_3_frame, alignment=ATOP, stretch=1)

        col_1 = QVBoxLayout()
        col_1.setSpacing(0)
        dmg_hider_label = self.create_label('Damage table columns:', 'label', col_1_frame)
        col_1.addWidget(dmg_hider_label)
        dmg_hider_layout = QVBoxLayout()
        dmg_hider_frame = self.create_frame(col_1_frame, style_override=
                {'border-color':'@lbg', 'border-width':'@bw', 'border-style':'solid', 'border-radius': 2})
        dmg_hider_frame.setSizePolicy(SMINMAX)
        for i, head in enumerate(TREE_HEADER[1:]):
            bt = self.create_button(head, 'toggle_button', dmg_hider_frame, toggle=True)
            bt.setSizePolicy(SMINMAX)
            bt.setChecked(self.settings.value(f'dmg_columns|{i}', type=bool))
            bt.clicked.connect(lambda state, i=i: self.settings.setValue(f'dmg_columns|{i}', state))
            dmg_hider_layout.addWidget(bt, stretch=1)
        dmg_hider_frame.setLayout(dmg_hider_layout)
        col_1.addWidget(dmg_hider_frame, alignment=ATOP)
        apply_button = self.create_button('Apply', 'button', col_1_frame, {'margin-top':15})
        apply_button.clicked.connect(self.update_shown_columns_dmg)
        col_1.addWidget(apply_button, alignment=ALEFT)
        col_1_frame.setLayout(col_1)

        col_2 = QVBoxLayout()
        col_2.setSpacing(0)
        heal_hider_label = self.create_label('Heal table columns:', 'label', col_2_frame)
        col_2.addWidget(heal_hider_label)
        heal_hider_layout = QVBoxLayout()
        heal_hider_frame = self.create_frame(col_2_frame, style_override=
                {'border-color':'@lbg', 'border-width':'@bw', 'border-style':'solid', 'border-radius': 2})
        heal_hider_frame.setSizePolicy(SMINMAX)
        for i, head in enumerate(HEAL_TREE_HEADER[1:]):
            bt = self.create_button(head, 'toggle_button', heal_hider_frame)
            bt.setCheckable(True)
            bt.setSizePolicy(SMINMAX)
            bt.setChecked(self.settings.value(f'heal_columns|{i}', type=bool))
            bt.clicked.connect(lambda state, i=i: self.settings.setValue(f'heal_columns|{i}', state))
            heal_hider_layout.addWidget(bt, stretch=1)
        heal_hider_frame.setLayout(heal_hider_layout)
        col_2.addWidget(heal_hider_frame, alignment=ATOP)
        apply_button_2 = self.create_button('Apply', 'button', col_2_frame, {'margin-top':15})
        apply_button_2.clicked.connect(self.update_shown_columns_heal)
        col_2.addWidget(apply_button_2, alignment=ALEFT)
        col_2_frame.setLayout(col_2)

        settings_frame.setLayout(settings_layout)

    def browse_log(self, entry:QLineEdit):
        """
        Callback for browse button.

        Parameters:
        - :param entry: QLineEdit -> path entry line widget
        """
        current_path = entry.text()
        if current_path != '':
            path = self.browse_path(os.path.dirname(current_path), 'Logfile (*.log);;Any File (*.*)')
            if path != '':
                entry.setText(format_path(path))

    def save_log(self, entry:QLineEdit):
        """
        Callback for save button.

        Parameters:
        - :param entry: QLineEdit -> path entry line widget
        """
        if self.parser1 and self.parser1.active_combat:
            current_path = entry.text()
            if current_path != '':
                path = self.browse_path(os.path.dirname(current_path), 'Logfile (*.log);;Any File (*.*)', 
                        save=True)
                if path != '':
                    with open(path, "w") as file:
                        for line in self.parser1.active_combat.log_data:
                            file.write(logline_to_str(line))

    def truncate_log(self, entry:QLineEdit):
        """
        Callback for truncate button.

        Parameters:
        - :param entry: QLineEdit -> path entry line widget
        """
        if self.parser1 and self.parser1.active_combat:
            current_path = entry.text()
            with open(current_path, "w") as file:
                for line in self.parser1.active_combat.log_data:
                    file.write(logline_to_str(line))
            self.analyze_log_callback(path=self.entry.text())

    def set_variable(self, var_to_be_set, index, value):
        """
        Assigns value at index to variable
        """
        var_to_be_set[index] = value



from PySide6.QtWidgets import QWidget, QPushButton, QTabWidget, QFrame, QTreeView, QComboBox, QTableView
from PySide6.QtGui import QIcon, QPixmap, QPainter, QFont
from PySide6.QtCore import QRect, Slot
from pyqtgraph import AxisItem, PlotWidget, BarGraphItem
import numpy as np

from .widgetbuilder import SMINMIN

class FlipButton(QPushButton):
    """
    QPushButton with two sets of commands, texts and icons that alter on click.
    """
    def __init__(self, r_text, l_text, parent, *ar, **kw):
        super().__init__(r_text, parent, *ar, **kw)
        self._r = True
        self._r_text = r_text
        self._l_text = l_text
        self.setText(r_text)
        self._r_function = self._f
        self._l_function = self._f
        self._r_icon = None
        self._l_icon = None
        self.clicked.connect(self.flip)

    @Slot()
    def flip(self):
        if self._r:
            self._r_function()
            self.setIcon(self._l_icon)
            self.setText(self._l_text)
            self._r = not self._r
        else:
            self._l_function()
            self.setIcon(self._r_icon)
            self.setText(self._r_text)
            self._r = not self._r

    def set_icon_r(self, icon:QIcon):
        self._r_icon = icon
        if self._r:
            self.setIcon(icon)

    def set_icon_l(self, icon:QIcon):
        self._l_icon = icon
        if not self._r:
            self.setIcon(icon)

    def set_text_r(self, text):
        self._r_text = text
        if self._r:
            self.setText(text)

    def set_text_l(self, text):
        self._l_text = text
        if not self._r:
            self.setText(text)

    def set_func_r(self, func):
        self._r_function = func

    def set_func_l(self, func):
        self._l_function = func
            
    def configure(self, settings):
        self.set_icon_r(settings['icon_r'])
        self.set_icon_l(settings['icon_l'])
        self.set_func_r(settings['func_r'])
        self.set_func_l(settings['func_l'])

    def _f(self):
        return

class BannerLabel(QWidget):
    """
    Label displaying image that resizes according to its parents width while preserving aspect ratio.
    """
    def __init__(self, path, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.setPixmap(QPixmap(path))
        self.setSizePolicy(SMINMIN)
        self.setMinimumHeight(10) # forces visibility

    def setPixmap(self, p):
        self.p = p
        self.update()

    def paintEvent(self, event):
        if not self.p.isNull():
            painter = QPainter(self)
            painter.setRenderHint(QPainter.RenderHint.SmoothPixmapTransform)
            w = int(self.rect().width())
            h = int(w * 126/2880)
            rect = QRect(0, 0, w, h)
            painter.drawPixmap(rect, self.p)
            self.setMaximumHeight(h)
            self.setMinimumHeight(h)

class CustomPlotAxis(AxisItem):
    """
    Extending AxisItem for custom tick formatting
    """
    def __init__(self, *args, unit: str = '', **kwargs):
        super().__init__(*args, **kwargs)
        self._unit = ' ' + unit

    @property
    def unit(self):
        return self._unit
    
    @unit.setter
    def unit(self, value):
        self._unit = ' ' + value

    def tickStrings(self, values, scale, spacing):
        if self.logMode:
            return self.logTickStrings(values, scale, spacing)
        
        strings = list()
        for tick in values:
            if tick >= 1000000:
                strings.append(f'{tick / 1000000:.2f} M{self._unit}')
            elif tick >= 1000:
                strings.append(f'{tick / 1000:.0f} k{self._unit}')
            else:
                strings.append(f'{tick:.0f}{self._unit}')
        return strings
    
class AnalysisPlot(PlotWidget):
    """
    PlotWidget for plotting the analysis plot.
    """
    def __init__(self, colors: tuple, tick_color: str, tick_font: QFont, legend_layout):
        """
        Parameters:
        - :param colors: tuple with at least 5 different colors that are used to paint the bars
        - :param tick_color: color for the tick annotations
        - :param tick_font: font for the tick annotations
        """
        super().__init__()
        self._bar_queue = list()
        self._legend_queue = list()
        self._bar_position = 0
        self._colors = colors
        self._frozen = True
        self._legend_layout = legend_layout
        left_axis = CustomPlotAxis('left')
        left_axis.setTickFont(tick_font)
        left_axis.setTextPen(color=tick_color)
        bottom_axis = CustomPlotAxis('bottom', unit='s')
        bottom_axis.setTickFont(tick_font)
        bottom_axis.setTextPen(color=tick_color)
        self.setAxisItems({'left': left_axis, 'bottom': bottom_axis})
        self.setBackground(None)
        self.setMouseEnabled(False, False)
        self.setMenuEnabled(False)
        self.hideButtons()
        self.setDefaultPadding(padding=0)

    def add_bar(self, data):
        """
        Adds plot item to plot widget and removes plot item if there are more than 5 currently displayed.

        Parameters:
        - :param data: list or array containing the height of the bars

        :return: returns the color that the graph was created with for the legend
        """
        if self._frozen:
            return
        time_reference = np.arange(len(data))
        group_width = 0.9
        bar_width = group_width / 5
        bar_offset = - (group_width / 2) + 0.5 * bar_width + self._bar_position * bar_width
        time_data = np.subtract(time_reference, bar_offset)
        brush_color = self._colors[self._bar_position]
        bars = BarGraphItem(x=time_data, width=bar_width, height=data, brush=brush_color, pen=None)
        if len(self._bar_queue) >= 5:
            self.removeItem(self._bar_queue.pop(0))
            self._legend_layout.removeWidget(self._legend_queue.pop(0))
        self._bar_queue.append(bars)
        self.addItem(bars)
        self._bar_position += 1
        if self._bar_position >=5:
            self._bar_position = 0
        return brush_color
    
    def add_legend_item(self, legend_item: QFrame):
        self._legend_queue.append(legend_item)
        self._legend_layout.addWidget(legend_item)

    def clear_plot(self):
        """
        Removes all bars from the plot
        """
        for bar in self._bar_queue:
            self.removeItem(bar)
        self._bar_queue = list()
        for legend_item in self._legend_queue:
            self._legend_layout.removeWidget(legend_item)
        self._legend_queue = list()
        self._bar_position = 0
    
    def toggle_freeze(self):
        """
        Freezes when unfrozen, unfreezes when frozen
        """
        self._frozen = not self._frozen

class WidgetStorage():
    """
    Class to store widgets.
    """
    def __init__(self):
        self.main_menu_buttons: list[QPushButton] = list()
        self.main_tabber: QTabWidget
        self.main_tab_frames: list[QFrame] = list()

        self.overview_menu_buttons: list[QPushButton] = list()
        self.overview_tabber: QTabWidget
        self.overview_tab_frames: list[QFrame] = list()
        
        self.analysis_menu_buttons: list[QPushButton] = list()
        self.analysis_tabber: QTabWidget
        self.analysis_tab_frames: list[QFrame] = list()
        self.analysis_table_dout: QTreeView
        self.analysis_table_dtaken: QTreeView
        self.analysis_table_hout: QTreeView
        self.analysis_table_hin: QTreeView
        self.analysis_plot_dout: AnalysisPlot
        self.analysis_plot_dtaken: AnalysisPlot
        self.analysis_plot_hout: AnalysisPlot
        self.analysis_plot_hin: AnalysisPlot

        self.ladder_map: QComboBox
        self.ladder_table: QTableView
